from typing import Union

from bluer_options.env.functions import load_config, load_env, get_env

load_config(__name__)


ABCUL = " \\\n\t"


bluer_ai_log_filename = get_env("bluer_ai_log_filename")


CYAN = "\033[36m"
GREEN = "\033[32m"
LIGHTBLUE = "\033[96m"
NC = "\033[0m"
RED = "\033[31m"

EOP = "\033[33m"

HOST_NAME: Union[None, str] = None

abcli_hostname = get_env("abcli_hostname")

BLUER_AI_WIFI_SSID = get_env("BLUER_AI_WIFI_SSID")

BLUER_OPTIONS_TIMEZONE = get_env("BLUER_OPTIONS_TIMEZONE")



import argparse
import random

from blueness import module

from bluer_options import NAME
from bluer_options import string

NAME = module.name(__file__, NAME)

list_of_tasks = "after | before | pretty_date | random"


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help=list_of_tasks,
)
parser.add_argument(
    "--include_time",
    type=int,
    help="0|1",
    default=1,
)
parser.add_argument(
    "--length",
    type=int,
    default=16,
)
parser.add_argument(
    "--string",
    type=str,
)
parser.add_argument(
    "--substring",
    type=str,
)
parser.add_argument(
    "--unique",
    type=int,
    help="0|1",
    default=0,
)
parser.add_argument(
    "--float",
    type=int,
    help="0|1",
    default=0,
)
parser.add_argument(
    "--min",
    type=float,
    default=0.0,
)
parser.add_argument(
    "--max",
    type=float,
    default=100.0,
)
args = parser.parse_args()

success = args.task in list_of_tasks.split(" | ")
if args.task == "after":
    print(string.after(args.string, args.substring))
elif args.task == "before":
    print(string.before(args.string, args.substring))
elif args.task == "pretty_date":
    print(
        string.pretty_date(
            as_filename=True,
            include_time=args.include_time == 1,
            unique=args.unique == 1,
        )
    )
elif args.task == "random":
    print(
        "{:8}".format(
            random.uniform(
                args.min,
                args.max,
            )
        )
        if args.float == 1
        else string.random(args.length)
    )
else:
    print(f"-{NAME}: {args.task}: command not found")

if not success:
    print(f"-{NAME}: {args.task}: failed")



import logging
from logging import Logger
from logging.handlers import RotatingFileHandler
from bluer_options.env import bluer_ai_log_filename

# to hide "botocore.credentials Found credentials in environment variables."
logging.getLogger("botocore.credentials").setLevel(logging.WARNING)


# Based on https://stackoverflow.com/a/22313803
logging.addLevelName(logging.INFO, "")
logging.addLevelName(logging.DEBUG, "❓ ")
logging.addLevelName(logging.ERROR, "❗️ ")
logging.addLevelName(logging.WARNING, "⚠️  ")

logging_level = logging.INFO

logging.getLogger().setLevel(logging_level)

log_formatter = logging.Formatter("%(levelname)s%(name)s %(message)s")
try:
    file_handler = RotatingFileHandler(
        bluer_ai_log_filename,
        maxBytes=10485760,
        backupCount=10000,
    )
    file_handler.setLevel(logging_level)
    file_handler.setFormatter(log_formatter)
    logging.getLogger().addHandler(file_handler)
except:
    pass

console_handler = logging.StreamHandler()
console_handler.setLevel(logging_level)
console_handler.setFormatter(log_formatter)
logging.getLogger().addHandler(console_handler)


def get_logger(ICON) -> Logger:
    return logging.getLogger(f"{ICON} ")


logger = get_logger("::")


# https://stackoverflow.com/a/10645855
def crash_report(description):
    logger.error(f"crash: {description}", exc_info=1)


def log_long_text(
    logger: Logger,
    text: str,
    max_length: int = 100,
):
    logger.info(
        "{:,} char(s): {}{}".format(
            len(text),
            text[:max_length],
            "..." if len(text) > max_length else "",
        )
    )


def log_dict(
    logger: Logger,
    title: str,
    dict_of_items: dict,
    item_name_plural: str = "item(s)",
    max_count: int = 5,
    max_length: int = 100,
):
    logger.info(f"{title} {len(dict_of_items)} {item_name_plural}.")

    for index, (item, info) in enumerate(dict_of_items.items()):
        logger.info(
            "#{: 4} - {}: {}{}".format(
                index,
                item,
                info[:max_length],
                "..." if len(info) > max_length else "",
            )
        )

        if index > max_count:
            logger.info("...")
            break


def log_list_as_str(
    title: str,
    list_of_items: list,
    item_name_plural: str = "item(s)",
    max_count: int = 5,
) -> str:
    return "{} {} {}: {}".format(
        title,
        len(list_of_items),
        item_name_plural,
        ", ".join(
            list_of_items
            if max_count == -1
            else (
                list_of_items[:max_count]
                + (["..."] if len(list_of_items) > max_count else [])
            )
        ),
    )


def log_list(
    logger: Logger,
    title: str,
    list_of_items: list,
    item_name_plural: str = "item(s)",
    max_count: int = 5,
    max_length: int = 100,
    itemize: bool = True,
):
    if not itemize:
        logger.info(
            log_list_as_str(
                title=title,
                list_of_items=list_of_items,
                item_name_plural=item_name_plural,
                max_count=max_count,
            )
        )
        return

    logger.info(f"{title} {len(list_of_items)} {item_name_plural}")

    for index, item in enumerate(list_of_items):
        logger.info(
            "#{: 4} - {}{}".format(
                index,
                item if max_length == -1 else item[:max_length],
                "" if max_length == -1 else "..." if len(item) > max_length else "",
            )
        )

        if max_count != -1 and index >= max_count - 1:
            logger.info("...")
            break



import time

from bluer_options.env import bluer_ai_log_filename


def watch(seconds: int = 1) -> bool:
    try:
        with open(bluer_ai_log_filename, "r") as f:
            f.seek(0, 2)  # jump to end of file
            while True:
                line = f.readline()
                if not line:
                    time.sleep(seconds)
                    continue
                print(line, end="")
    except KeyboardInterrupt:
        print("\n^C received.")

    return True



from typing import Any
import os
from dotenv import load_dotenv

# https://chatgpt.com/c/683d39ac-34a8-8005-b780-71a6d2253ea9
try:
    from importlib.resources import files, as_file
except ImportError:
    # for Python < 3.9
    from importlib_resources import files, as_file


def get_env(name: str, default: Any = "") -> Any:
    output = os.getenv(
        name,
        default,
    )

    if isinstance(default, bool):
        return output == "1"

    # order is critical
    for output_type in [int, float]:
        if isinstance(default, output_type):
            try:
                return output_type(output)
            except:
                return default

    return output


def load_config(
    package_name: str,
    verbose: bool = False,
):
    if "." in package_name:
        package_name = package_name.split(".")[0]

    resource = files(package_name).joinpath("config.env")
    with as_file(resource) as env_filename:
        if verbose:
            print(f"loading {env_filename}.")

        assert load_dotenv(
            env_filename
        ), f"package_name: {package_name}, env_filename: {env_filename}"


def load_env(
    package_name: str,
    verbose: bool = False,
):
    if "." in package_name:
        package_name = package_name.split(".")[0]

    with as_file(files(package_name)) as package_path:
        env_filename = os.path.join(
            os.path.dirname(package_path),
            ".env",
        )

        if verbose:
            print(f"loading {env_filename}.")

        load_dotenv(env_filename)



from bluer_options.host.functions import (
    get_name_,
    get_name,
    get_seed_filename,
    is_64bit,
    is_aws_batch,
    is_docker,
    is_ec2,
    is_github_workflow,
    is_headless,
    is_jetson,
    is_jupyter,
    is_mac,
    is_rpi,
    is_ssh_session,
    is_ubuntu,
    signature,
    tensor_processing_signature,
)



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_options import NAME
from bluer_options.logger.watch import watch
from bluer_options.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="watch",
)
parser.add_argument(
    "--seconds",
    type=int,
    default=1,
    help="in seconds.",
)
args = parser.parse_args()

success = False
if args.task == "watch":
    success = watch(seconds=args.seconds)
else:
    success = None

sys_exit(logger, NAME, args.task, success)



import time
from bluer_options import string


class ElapsedTimer:
    def __init__(self):
        self.start_time = time.time()
        self.elapsed_time_ = None

    def as_str(
        self,
        stop: bool = True,
        **kwargs,
    ):
        if stop:
            self.stop()

        elapsed_time = self.elapsed_time

        return (
            "None"
            if elapsed_time is None
            else string.pretty_duration(elapsed_time, **kwargs)
        )

    @property
    def elapsed_time(self) -> float:
        if self.start_time is None:
            return self.elapsed_time_

        return time.time() - self.start_time

    def reset(self):
        self.start_time = time.time()
        self.elapsed_time_ = None

    def stop(self):
        if self.start_time is None:
            return

        self.elapsed_time_ = self.elapsed_time
        self.start_time = None



from bluer_options import NAME, VERSION, DESCRIPTION, REPO_NAME
from blueness.pypi import setup


setup(
    filename=__file__,
    repo_name=REPO_NAME,
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    packages=[
        NAME,
        f"{NAME}.assets",
        f"{NAME}.env",
        f"{NAME}.help",
        f"{NAME}.host",
        f"{NAME}.list",
        f"{NAME}.logger",
        f"{NAME}.options",
        f"{NAME}.string",
        f"{NAME}.terminal",
        f"{NAME}.timing",
    ],
    include_package_data=True,
    package_data={
        NAME: [
            "config.env",
            ".bash/**/*.sh",
            "assets/*",
        ],
    },
)



import os
from typing import List

list_of_modules: List[str] = [
    "abcli",
    "bluer_ai",
    "bluer_algo",
    "bluer_flow",
    "bluer_geo",
    "bluer_journal",
    "bluer_options",
    "bluer_objects",
    "bluer_plugin",
    "bluer_sandbox",
    "bluer_sbc",
    "bluer_ugv",
    "giza",
] + [
    item
    for item in os.getenv(
        "BLUE_OPTIONS_HELP_MODULE_LIST",
        "",
    ).split("+")
    if item
]


def get_callable_module(
    callable: str,
    module_name_check: bool = True,
) -> str:
    for module in list_of_modules:
        if callable.startswith(module):
            return (
                os.getenv(f"{module}_module_name", module)
                if module_name_check
                else module
            )

    return callable


def get_callable_suffix(callable: str) -> str:
    module = get_callable_module(
        callable,
        module_name_check=False,
    )

    suffix = callable.split(module)[1]

    if suffix.startswith("_"):
        suffix = suffix[1:]

    return suffix



import time
from typing import List, Dict
from collections import defaultdict
from functools import wraps

from bluer_options import string
from bluer_options.logger import logger
from bluer_options.logger.config import log_list
from bluer_options.timing.elapsed_timer import ElapsedTimer


def recursive_to_dict(d):
    if isinstance(d, (defaultdict, dict)):
        d = {k: recursive_to_dict(v) for k, v in d.items()}
    return d


class Timing:
    def __init__(self):
        self._active = {}
        self.stats = defaultdict(lambda: {"count": 0, "total": 0.0})
        self.elapsed_timer = ElapsedTimer()

    def start(self, keyword: str):
        """Start timing a code block identified by 'keyword'."""
        self._active[keyword] = time.perf_counter()

    def stop(self, keyword: str):
        """Stop timing for 'keyword' and record elapsed time."""
        if keyword not in self._active:
            raise RuntimeError(f"No active timer for {keyword}")

        elapsed = time.perf_counter() - self._active.pop(keyword)
        self.stats[keyword]["count"] += 1
        self.stats[keyword]["total"] += elapsed

        return elapsed

    @property
    def as_dict(self) -> Dict:
        self.calculate()

        return recursive_to_dict(self.stats)

    def as_list(self, **kwgrs) -> List[str]:
        self.calculate()

        lines = []
        # sort items by total time (descending)
        for k, v in sorted(
            self.stats.items(),
            key=lambda item: item[1]["total"],
            reverse=True,
        ):
            lines.append(
                "{}: called {:,} time(s), total {}, avg {}".format(
                    k,
                    v["count"],
                    string.pretty_duration(v["total"], **kwgrs),
                    string.pretty_duration(v["average"], **kwgrs),
                )
            )
        return lines

    def calculate(self):
        for v in self.stats.values():
            v["average"] = v["total"] / v["count"]

    def log(
        self,
        include_ms: bool = True,
        largest: bool = True,
        short: bool = True,
        **kwargs,
    ):
        log_list(
            logger,
            "in {} called".format(
                self.elapsed_timer.as_str(
                    stop=False,
                    include_ms=include_ms,
                    largest=largest,
                    short=short,
                )
            ),
            self.as_list(
                include_ms=include_ms,
                largest=largest,
                short=short,
            ),
            "function(s):",
            max_count=-1,
            **kwargs,
        )

    def reset(self):
        self._active = {}
        self.stats = defaultdict(lambda: {"count": 0, "total": 0.0})
        self.elapsed_timer.reset()

    def time(self, arg=None):
        """Use as @timing.time, @timing.time(), or @timing.time('custom')"""

        # Case 1: @timing.time  (no parentheses, arg is the function)
        if callable(arg):
            func = arg
            name = func.__name__

            @wraps(func)
            def inner(*args, **kwargs):
                self.start(name)
                try:
                    return func(*args, **kwargs)
                finally:
                    self.stop(name)

            return inner

        # Case 2: @timing.time() or @timing.time("custom")
        keyword = arg  # may be None or str

        def decorator(func):
            name = keyword or func.__name__

            @wraps(func)
            def inner(*args, **kwargs):
                self.start(name)
                try:
                    return func(*args, **kwargs)
                finally:
                    self.stop(name)

            return inner

        return decorator
